=== modified file 'NEWS'
--- NEWS	2007-09-25 14:36:11 +0000
+++ NEWS	2007-09-26 19:51:30 +0000
@@ -1,5 +1,23 @@
 0.92.0	UNRELEASED
 
+ FEATURES
+ 
+  * Rewrite of a large ammount of the viz code:
+        * Make use of the bzrlib.tsort.merge_sort for sorting the revisions.
+        * A new algorithm to work out how to draw the lines using the
+          revsion numbers returned from merge_sort.
+        * Lines that are very long are broken, which results in the graph
+          requiring less space to render and the graph is easier to interpret.
+          (#82907)
+        * Choose the color for a branch from the revsion number, which results
+          in stable colors. Revisions on the mainline are allways black.
+        * Implement a TreeModel that loads the revision data only when a a
+          revision is scrolled into view, improving the time it takes to display
+          the first screen of data by 92%. (#124143)
+        * Add a column to display the revision number. (#64167)
+        * Add buttons to navigate to child revisions.
+    (Gary van der Merwe)
+    
  INTERNALS
 
   * Use standard version tuple. (Jelmer)
@@ -12,7 +30,7 @@
 
   * Added keyboard shortcuts for closing windows and exiting Olive.
     (#139559, Daniel Schierbeck)
-  
+ 
   * The tip revision will be automatically selected in the branch visualization
     window. (#144244, Daniel Schierbeck)
   
@@ -22,8 +40,8 @@
  BUG FIXES
 
   * Fix the "Complete Diff" entry in DiffWindow to show the whole diff.
-    (#140512, James Westby)
-  
+    (#140512, James Westby) 
+ 
   * Right-clicking in Olive's file list works again. (#136310, Mark Lee)
   
   * Annotate context menu works again. (Daniel Schierbeck)
@@ -37,7 +55,7 @@
  INTERNALS
  
   * `bzr viz --limit` much more faster than before. (John Arbash Meinel)
-  
+
   * Use get_apparent_author, rename variables to 'author'. (Aaron Bentley)
   
   * Added default_branch_path to MergeDialog and use it from olive-gtk.

=== modified file 'logview.py'
--- logview.py	2007-08-16 14:52:00 +0000
+++ logview.py	2007-08-31 11:24:56 +0000
@@ -30,13 +30,15 @@
     start.
     """
 
-    def __init__(self, revision=None, scroll=True, tags=[]):
+    def __init__(self, revision=None, scroll=True, tags=[],
+                 show_children=False):
         super(LogView, self).__init__()
         if scroll:
             self.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
         else:
             self.set_policy(gtk.POLICY_NEVER, gtk.POLICY_NEVER)
         self.set_shadow_type(gtk.SHADOW_NONE)
+        self.show_children = show_children
         self._create()
         self._show_callback = None
         self._go_callback = None
@@ -51,7 +53,7 @@
     def set_go_callback(self, callback):
         self._go_callback = callback
 
-    def set_revision(self, revision, tags=[]):
+    def set_revision(self, revision, tags=[], children=[]):
         self._revision = revision
         self.revision_id.set_text(revision.revision_id)
         if revision.committer is not None:
@@ -76,7 +78,15 @@
         except KeyError:
             self.branchnick_label.set_text("")
 
-        self._add_parents(revision.parent_ids)
+        self._add_parents_or_children(revision.parent_ids,
+                                      self.parents_widgets,
+                                      self.parents_table)
+        
+        if self.show_children:
+            self._add_parents_or_children(children,
+                                          self.children_widgets,
+                                          self.children_table)
+        
         self._add_tags(tags)
 
     def _show_clicked_cb(self, widget, revid, parentid):
@@ -107,17 +117,17 @@
         self.tags_label.show_all()
         
 
-    def _add_parents(self, parent_ids):
-        for widget in self.parents_widgets:
-            self.parents_table.remove(widget)
-            
-        self.parents_widgets = []
-        self.parents_table.resize(max(len(parent_ids), 1), 2)
+    def _add_parents_or_children(self, revids, widgets, table):
+        while len(widgets) > 0:
+            widget = widgets.pop()
+            table.remove(widget)
+        
+        table.resize(max(len(revids), 1), 2)
 
-        for idx, parent_id in enumerate(parent_ids):
+        for idx, revid in enumerate(revids):
             align = gtk.Alignment(0.0, 0.0)
-            self.parents_widgets.append(align)
-            self.parents_table.attach(align, 1, 2, idx, idx + 1,
+            widgets.append(align)
+            table.attach(align, 1, 2, idx, idx + 1,
                                       gtk.EXPAND | gtk.FILL, gtk.FILL)
             align.show()
 
@@ -134,15 +144,15 @@
                 button = gtk.Button()
                 button.add(image)
                 button.connect("clicked", self._show_clicked_cb,
-                               self._revision.revision_id, parent_id)
+                               self._revision.revision_id, revid)
                 hbox.pack_start(button, expand=False, fill=True)
                 button.show()
 
             if self._go_callback is not None:
-                button = gtk.Button(parent_id)
-                button.connect("clicked", self._go_clicked_cb, parent_id)
+                button = gtk.Button(revid)
+                button.connect("clicked", self._go_clicked_cb, revid)
             else:
-                button = gtk.Label(parent_id)
+                button = gtk.Label(revid)
             button.set_use_underline(False)
             hbox.pack_start(button, expand=False, fill=True)
             button.show()
@@ -151,7 +161,7 @@
         vbox = gtk.VBox(False, 6)
         vbox.set_border_width(6)
         vbox.pack_start(self._create_headers(), expand=False, fill=True)
-        vbox.pack_start(self._create_parents_table(), expand=False, fill=True)
+        vbox.pack_start(self._create_parents_and_children(), expand=False, fill=True)
         vbox.pack_start(self._create_message_view())
         self.add_with_viewport(vbox)
         vbox.show()
@@ -261,22 +271,41 @@
 
         return self.table
 
-    def _create_parents_table(self):
-        self.parents_table = gtk.Table(rows=1, columns=2)
-        self.parents_table.set_row_spacings(3)
-        self.parents_table.set_col_spacings(6)
-        self.parents_table.show()
+    
+    def _create_parents_and_children(self):
+        hbox = gtk.HBox(True, 6)
+        
+        self.parents_table = self._create_parents_or_children_table(
+            "<b>Parents:</b>")
         self.parents_widgets = []
+        hbox.pack_start(self.parents_table)
+        
+        if self.show_children:
+            self.children_table = self._create_parents_or_children_table(
+                "<b>Children:</b>")
+            self.children_widgets = []
+            hbox.pack_start(self.children_table)
+        
+        hbox.show()
+        return hbox
+        
+    def _create_parents_or_children_table(self, text):
+        table = gtk.Table(rows=1, columns=2)
+        table.set_row_spacings(3)
+        table.set_col_spacings(6)
+        table.show()
 
         label = gtk.Label()
-        label.set_markup("<b>Parents:</b>")
+        label.set_markup(text)
         align = gtk.Alignment(0.0, 0.5)
         align.add(label)
-        self.parents_table.attach(align, 0, 1, 0, 1, gtk.FILL, gtk.FILL)
+        table.attach(align, 0, 1, 0, 1, gtk.FILL, gtk.FILL)
         label.show()
         align.show()
 
-        return self.parents_table
+        return table
+    
+
 
     def _create_message_view(self):
         self.message_buffer = gtk.TextBuffer()

=== modified file 'viz/branchwin.py'
--- viz/branchwin.py	2007-09-23 18:32:58 +0000
+++ viz/branchwin.py	2007-09-26 18:54:49 +0000
@@ -12,11 +12,13 @@
 import gtk
 import gobject
 import pango
+import treemodel
 
 from bzrlib.osutils import format_date
 
-from graph import distances, graph, same_branch
+from linegraph import linegraph, same_branch
 from graphcell import CellRendererGraph
+from treemodel import TreeModel
 
 
 class BranchWindow(gtk.Window):
@@ -81,25 +83,40 @@
         self.treeview.connect("row-activated", self._treeview_row_activated_cb)
         self.treeview.connect("button-release-event", 
                 self._treeview_row_mouseclick)
+        self.treeview.set_property('fixed-height-mode', True)
         scrollwin.add(self.treeview)
         self.treeview.show()
 
-        cell = CellRendererGraph()
-        column = gtk.TreeViewColumn()
+        cell = gtk.CellRendererText()
+        cell.set_property("width-chars", 15)
+        cell.set_property("ellipsize", pango.ELLIPSIZE_END)
+        column = gtk.TreeViewColumn("Revision No")
         column.set_resizable(True)
-        column.pack_start(cell, expand=False)
-        column.add_attribute(cell, "node", 1)
-        column.add_attribute(cell, "in-lines", 2)
-        column.add_attribute(cell, "out-lines", 3)
+        column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
+        column.set_fixed_width(cell.get_size(self.treeview)[2])
+        column.pack_start(cell, expand=True)
+        column.add_attribute(cell, "text", treemodel.REVNO)
         self.treeview.append_column(column)
 
+        self.graph_cell = CellRendererGraph()
+        self.graph_column = gtk.TreeViewColumn()
+        self.graph_column.set_resizable(True)
+        self.graph_column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
+        self.graph_column.pack_start(self.graph_cell, expand=False)
+        self.graph_column.add_attribute(self.graph_cell, "node", treemodel.NODE)
+        self.graph_column.add_attribute(self.graph_cell, "in-lines", treemodel.LAST_LINES)
+        self.graph_column.add_attribute(self.graph_cell, "out-lines", treemodel.LINES)
+        self.treeview.append_column(self.graph_column)
+
         cell = gtk.CellRendererText()
         cell.set_property("width-chars", 40)
         cell.set_property("ellipsize", pango.ELLIPSIZE_END)
         column = gtk.TreeViewColumn("Message")
         column.set_resizable(True)
+        column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
+        column.set_fixed_width(cell.get_size(self.treeview)[2])
         column.pack_start(cell, expand=True)
-        column.add_attribute(cell, "text", 4)
+        column.add_attribute(cell, "text", treemodel.MESSAGE)
         self.treeview.append_column(column)
 
         cell = gtk.CellRendererText()
@@ -107,16 +124,21 @@
         cell.set_property("ellipsize", pango.ELLIPSIZE_END)
         column = gtk.TreeViewColumn("Committer")
         column.set_resizable(True)
+        column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
+        column.set_fixed_width(cell.get_size(self.treeview)[2])
         column.pack_start(cell, expand=True)
-        column.add_attribute(cell, "text", 5)
+        column.add_attribute(cell, "text", treemodel.COMMITER)
         self.treeview.append_column(column)
 
         cell = gtk.CellRendererText()
+        cell.set_property("width-chars", 40)
         cell.set_property("ellipsize", pango.ELLIPSIZE_END)
         column = gtk.TreeViewColumn("Date")
         column.set_resizable(True)
+        column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
+        column.set_fixed_width(cell.get_size(self.treeview)[2])
         column.pack_start(cell, expand=True)
-        column.add_attribute(cell, "text", 6)
+        column.add_attribute(cell, "text", treemodel.TIMESTAMP)
         self.treeview.append_column(column)
 
         return scrollwin
@@ -152,7 +174,7 @@
     def construct_bottom(self):
         """Construct the bottom half of the window."""
         from bzrlib.plugins.gtk.logview import LogView
-        self.logview = LogView()
+        self.logview = LogView(None, True, [], True)
         (width, height) = self.get_size()
         self.logview.set_size_request(width, int(height / 2.5))
         self.logview.show()
@@ -168,40 +190,19 @@
         treeview itself.
         """
         self.branch = branch
-
-        # [ revision, node, last_lines, lines, message, committer, timestamp ]
-        self.model = gtk.ListStore(gobject.TYPE_PYOBJECT,
-                                   gobject.TYPE_PYOBJECT,
-                                   gobject.TYPE_PYOBJECT,
-                                   gobject.TYPE_PYOBJECT,
-                                   str, str, str)
-        self.index = {}
         self.set_title(branch.nick + " - bzrk")
         gobject.idle_add(self.populate_model, start, maxnum)
 
     def populate_model(self, start, maxnum):
-        index = 0
-        
-        last_lines = []
-        (self.revisions, colours, self.children, self.parent_ids,
-            merge_sorted) = distances(self.branch.repository, start,
-                                      maxnum=maxnum)
-        for (index, (revision, node, lines)) in enumerate(graph(
-                self.revisions, colours, merge_sorted)):
-            # FIXME: at this point we should be able to show the graph order
-            # and lines with no message or commit data - and then incrementally
-            # fill the timestamp, committer etc data as desired.
-            message = revision.message.split("\n")[0]
-            if revision.committer is not None:
-                timestamp = format_date(revision.timestamp, revision.timezone)
-            else:
-                timestamp = None
-            self.model.append([revision, node, last_lines, lines,
-                               message, revision.committer, timestamp])
-            self.index[revision] = index
-            last_lines = lines
-            if maxnum is not None and index > maxnum:
-                break
+        (linegraphdata, index, columns_len) = linegraph(self.branch,
+                                                        start,
+                                                        maxnum)
+        self.model = TreeModel(self.branch, linegraphdata)
+        self.graph_cell.columns_len = columns_len
+        width = self.graph_cell.get_size(self.treeview)[2]
+        self.graph_column.set_fixed_width(width)
+        self.graph_column.set_max_width(width)
+        self.index = index
         self.treeview.set_model(self.model)
         self.treeview.get_selection().select_path(0)
         return False
@@ -222,58 +223,62 @@
     def _on_key_press_q(self, event):
         if event.state & gtk.gdk.CONTROL_MASK:
             gtk.main_quit()
-
+    
     def _treeselection_changed_cb(self, selection, *args):
         """Callback for when the treeview changes."""
         (model, selected_rows) = selection.get_selected_rows()
         if len(selected_rows) > 0:
-            revision = self.model[selected_rows[0]][0]
-
-            self.back_button.set_sensitive(len(self.parent_ids[revision]) > 0)
-            self.fwd_button.set_sensitive(len(self.children[revision]) > 0)
+            iter = self.model.get_iter(selected_rows[0])
+            revision = self.model.get_value(iter, treemodel.REVISION)
+            parents = self.model.get_value(iter, treemodel.PARENTS)
+            children = self.model.get_value(iter, treemodel.CHILDREN)
+            
+            self.back_button.set_sensitive(len(parents) > 0)
+            self.fwd_button.set_sensitive(len(children) > 0)
             tags = []
             if self.branch.supports_tags():
                 tagdict = self.branch.tags.get_reverse_tag_dict()
                 if tagdict.has_key(revision.revision_id):
                     tags = tagdict[revision.revision_id]
-            self.logview.set_revision(revision, tags)
+            self.logview.set_revision(revision, tags, children)
 
     def _back_clicked_cb(self, *args):
         """Callback for when the back button is clicked."""
         (path, col) = self.treeview.get_cursor()
         revision = self.model[path][0]
-        if not len(self.parent_ids[revision]):
+        parents = self.model[path][4]
+        if not len(parents):
             return
 
-        for parent_id in self.parent_ids[revision]:
-            parent = self.revisions[parent_id]
+        for parent_id in parents:
+            parent = self.revisions[self.index[parent_id]]
             if same_branch(revision, parent):
-                self.treeview.set_cursor(self.index[parent])
+                self.treeview.set_cursor(self.index[parent_id])
                 break
         else:
-            next = self.revisions[self.parent_ids[revision][0]]
-            self.treeview.set_cursor(self.index[next])
+            self.treeview.set_cursor(self.index[parents[0]])
         self.treeview.grab_focus()
 
     def _fwd_clicked_cb(self, *args):
         """Callback for when the forward button is clicked."""
         (path, col) = self.treeview.get_cursor()
         revision = self.model[path][0]
-        if not len(self.children[revision]):
+        children = self.model[path][5]
+        if not len(children):
             return
 
-        for child in self.children[revision]:
+        for child_id in children:
+            child = self.revisions[self.index[child_id]]
             if same_branch(child, revision):
-                self.treeview.set_cursor(self.index[child])
+                self.treeview.set_cursor(self.index[child_id])
                 break
         else:
-            prev = list(self.children[revision])[0]
-            self.treeview.set_cursor(self.index[prev])
+            self.treeview.set_cursor(self.index[children[0]])
         self.treeview.grab_focus()
 
     def _go_clicked_cb(self, revid):
         """Callback for when the go button for a parent is clicked."""
-        self.treeview.set_cursor(self.index[self.revisions[revid]])
+        self.treeview.set_cursor(self.index[revid])
         self.treeview.grab_focus()
 
     def show_diff(self, branch, revid, parentid):
@@ -300,7 +305,7 @@
             menu.popup(None, None, None, event.button, event.get_time())
 
     def selected_revision(self, path):
-        return self.model[path][0]
+        return self.model[path][treemodel.REVISION]
 
     def selected_revisions(self):
         return [self.selected_revision(path) for path in \

=== modified file 'viz/graphcell.py'
--- viz/graphcell.py	2006-12-21 06:19:52 +0000
+++ viz/graphcell.py	2007-09-22 15:01:53 +0000
@@ -30,6 +30,8 @@
       in_lines          (start, end, colour) tuple list to draw inward lines,
       out_lines         (start, end, colour) tuple list to draw outward lines.
     """
+    
+    columns_len = 0
 
     __gproperties__ = {
         "node":         ( gobject.TYPE_PYOBJECT, "node",
@@ -45,7 +47,7 @@
                           gobject.PARAM_WRITABLE
                         ),
         }
-
+    
     def do_set_property(self, property, value):
         """Set properties from GObject properties."""
         if property.name == "node":
@@ -84,6 +86,7 @@
         colours and the fg parameter provides the multiplier that should be
         applied to the foreground colours.
         """
+        mainline_color = ( 0.0, 0.0, 0.0 )
         colours = [
             ( 1.0, 0.0, 0.0 ),
             ( 1.0, 1.0, 0.0 ),
@@ -93,10 +96,14 @@
             ( 1.0, 0.0, 1.0 ),
             ]
 
-        colour %= len(colours)
-        red   = (colours[colour][0] * fg) or bg
-        green = (colours[colour][1] * fg) or bg
-        blue  = (colours[colour][2] * fg) or bg
+        if colour == 0:
+            colour_rgb = mainline_color
+        else:
+            colour_rgb = colours[colour % len(colours)]
+
+        red   = (colour_rgb[0] * fg) or bg
+        green = (colour_rgb[1] * fg) or bg
+        blue  = (colour_rgb[2] * fg) or bg
 
         ctx.set_source_rgb(red, green, blue)
 
@@ -109,11 +116,7 @@
         """
         box_size = self.box_size(widget) + 1
 
-        cols = self.node[0]
-        for start, end, colour in self.in_lines + self.out_lines:
-            cols = max(cols, start, end)
-
-        width = box_size * (cols + 1)
+        width = box_size * (self.columns_len + 1)
         height = box_size
 
         # FIXME I have no idea how to use cell_area properly
@@ -140,48 +143,19 @@
         box_size = self.box_size(widget)
 
         ctx.set_line_width(box_size / 8)
-        ctx.set_line_cap(cairo.LINE_CAP_SQUARE)
+        ctx.set_line_cap(cairo.LINE_CAP_ROUND)
 
         # Draw lines into the cell
         for start, end, colour in self.in_lines:
-            ctx.move_to(cell_area.x + box_size * start + box_size / 2,
-                        bg_area.y - bg_area.height / 2)
-
-            if start - end > 1:
-                ctx.line_to(cell_area.x + box_size * start, bg_area.y)
-                ctx.line_to(cell_area.x + box_size * end + box_size, bg_area.y)
-            elif start - end < -1:
-                ctx.line_to(cell_area.x + box_size * start + box_size,
-                            bg_area.y)
-                ctx.line_to(cell_area.x + box_size * end, bg_area.y)
-
-            ctx.line_to(cell_area.x + box_size * end + box_size / 2,
-                        bg_area.y + bg_area.height / 2)
-
-            self.set_colour(ctx, colour, 0.0, 0.65)
-            ctx.stroke()
+            self.render_line (ctx, cell_area, box_size,
+                         bg_area.y, bg_area.height,
+                         start, end, colour)
 
         # Draw lines out of the cell
         for start, end, colour in self.out_lines:
-            ctx.move_to(cell_area.x + box_size * start + box_size / 2,
-                        bg_area.y + bg_area.height / 2)
-
-            if start - end > 1:
-                ctx.line_to(cell_area.x + box_size * start,
-                            bg_area.y + bg_area.height)
-                ctx.line_to(cell_area.x + box_size * end + box_size,
-                            bg_area.y + bg_area.height)
-            elif start - end < -1:
-                ctx.line_to(cell_area.x + box_size * start + box_size,
-                            bg_area.y + bg_area.height)
-                ctx.line_to(cell_area.x + box_size * end,
-                            bg_area.y + bg_area.height)
-
-            ctx.line_to(cell_area.x + box_size * end + box_size / 2,
-                        bg_area.y + bg_area.height / 2 + bg_area.height)
-
-            self.set_colour(ctx, colour, 0.0, 0.65)
-            ctx.stroke()
+            self.render_line (ctx, cell_area, box_size,
+                         bg_area.y + bg_area.height, bg_area.height,
+                         start, end, colour)
 
         # Draw the revision node in the right column
         (column, colour) = self.node
@@ -194,3 +168,38 @@
 
         self.set_colour(ctx, colour, 0.5, 1.0)
         ctx.fill()
+    
+    def render_line (self, ctx, cell_area, box_size, mid, height, start, end, colour):
+        if start is None:
+            x = cell_area.x + box_size * end + box_size / 2
+            ctx.move_to(x, mid + height / 3)
+            ctx.line_to(x, mid + height / 3)
+            ctx.move_to(x, mid + height / 6)
+            ctx.line_to(x, mid + height / 6)
+            
+        elif end is None:
+            x = cell_area.x + box_size * start + box_size / 2
+            ctx.move_to(x, mid - height / 3)
+            ctx.line_to(x, mid - height / 3)
+            ctx.move_to(x, mid - height / 6)
+            ctx.line_to(x, mid - height / 6)
+        else:
+            startx = cell_area.x + box_size * start + box_size / 2
+            endx = cell_area.x + box_size * end + box_size / 2
+            
+            ctx.move_to(startx, mid - height / 2)
+            
+            if start - end == 0 :
+                ctx.line_to(endx, mid + height / 2)
+            else:
+                ctx.curve_to(startx, mid - height / 5,
+                             startx, mid - height / 5,
+                             startx + (endx - startx) / 2, mid)
+                
+                ctx.curve_to(endx, mid + height / 5,
+                             endx, mid + height / 5 ,
+                             endx, mid + height / 2)
+                
+        self.set_colour(ctx, colour, 0.0, 0.65)
+        ctx.stroke()
+        
\ No newline at end of file

=== renamed file 'viz/graph.py' => 'viz/linegraph.py'
--- viz/graph.py	2007-09-21 17:42:04 +0000
+++ viz/linegraph.py	2007-09-24 09:47:47 +0000
@@ -1,460 +1,363 @@
-# -*- coding: UTF-8 -*-
-"""Directed graph production.
-
-This module contains the code to produce an ordered directed graph of a
-bzr branch, such as we display in the tree view at the top of the bzrk
-window.
-"""
-
-__copyright__ = "Copyright © 2005 Canonical Ltd."
-__author__    = "Scott James Remnant <scott@ubuntu.com>"
-
-
-from bzrlib.revision import Revision
-from bzrlib.tsort import merge_sort
-
-
-class DummyRevision(Revision):
-    """Dummy bzr revision.
-
-    Sometimes, especially in older bzr branches, a revision is referenced
-    as the parent of another but not actually present in the branch's store.
-    When this happens we use an instance of this class instead of the real
-    Revision object (which we can't get).
-    """
-    def __init__(self, revid):
-        super(DummyRevision, self).__init__(revid)
-        self.committer = None
-        self.timestamp = None
-        self.timezone = None
-        self.message = revid
-
-
-class RevisionProxy(object):
-    """A revision proxy object.
-
-    This will demand load the revision it represents when the committer or
-    message attributes are accessed in order to populate them. It is 
-    constructed with the revision id and parent_ids list and a repository
-    object to request the revision from when needed.
-    """
-
-    def __init__(self, revid, parent_ids, repository):
-        self.revision_id = revid
-        self.parent_ids = parent_ids
-        self._repository = repository
-        self._revision = None
-
-    def _get_attribute_getter(attr):
-        def get_attribute(self):
-            if self._revision is None:
-                self._load()
-            return getattr(self._revision, attr)
-        return get_attribute
-    committer = property(_get_attribute_getter('committer'))
-    message = property(_get_attribute_getter('message'))
-    properties = property(_get_attribute_getter('properties'))
-    timestamp = property(_get_attribute_getter('timestamp'))
-    timezone = property(_get_attribute_getter('timezone'))
-
-    def _load(self):
-        """Load the revision object."""
-        self._revision = self._repository.get_revision(self.revision_id)
-
-
-class DistanceMethod(object):
-
-    def __init__(self, repository, start_revid):
-        self.repository = repository
-        self.start_revid = start_revid
-        self.revisions = {}
-        self.children_of_id = {start_revid: set()}
-        self.parent_ids_of = {}
-        self.colours = { start_revid: 0 }
-        self.last_colour = 0
-        self.direct_parent_of = {}
-        self.graph = {}
-
-    def fill_caches(self):
-        graph = self.repository.get_revision_graph_with_ghosts([self.start_revid])
-        for revid in graph.ghosts:
-            self.cache_revision(DummyRevision(revid))
-        for revid, parents in graph.get_ancestors().items():
-            self.cache_revision(RevisionProxy(revid, parents, self.repository))
-
-    def cache_revision(self, revision):
-        "Set the caches for a newly retrieved revision."""
-        revid = revision.revision_id
-        # Build a revision cache
-        self.revisions[revid] = revision
-        # Build a children dictionary
-        for parent_id in revision.parent_ids:
-            self.children_of_id.setdefault(parent_id, set()).add(revision)
-        # Build a parents dictionnary, where redundant parents will be removed,
-        # and that will be passed along tothe rest of program.
-        if len(revision.parent_ids) != len(set(revision.parent_ids)):
-            # fix the parent_ids list.
-            parent_ids = []
-            parent_ids_set = set()
-            for parent_id in revision.parent_ids:
-                if parent_id in parent_ids_set:
-                    continue
-                parent_ids.append(parent_id)
-                parent_ids_set.add(parent_id)
-            revision.parent_ids = parent_ids
-        self.parent_ids_of[revision] = list(revision.parent_ids)
-        self.graph[revid] = revision.parent_ids
-
-    def make_children_map(self):
-        revisions = self.revisions
-        return dict((revisions[revid], c)
-                    for (revid, c) in self.children_of_id.iteritems())
-
-    def sort_revisions(self, sorted_revids, maxnum):
-        revisions = self.revisions
-        parent_ids_of = self.parent_ids_of
-        children_of_id = self.children_of_id
-        # Try to compact sequences of revisions on the same branch.
-        distances = {}
-        skipped_revids = []
-        expected_id = sorted_revids[0]
-        pending_ids = []
-        while True:
-            revid = sorted_revids.pop(0)
-            if revid != expected_id:
-                skipped_revids.append(revid)
-                continue
-            revision = revisions[revid]
-            for child in children_of_id[revid]:
-                # postpone if any child is missing
-                if child.revision_id not in distances:
-                    if expected_id not in pending_ids:
-                        pending_ids.append(expected_id)
-                    expected_id = pending_ids.pop(0)
-                    skipped_revids.append(revid)
-                    sorted_revids[:0] = skipped_revids
-                    del skipped_revids[:]
-                    break
-            else:
-                # all children are here, push!
-                distances[revid] = len(distances)
-                if maxnum is not None and len(distances) > maxnum:
-                    # bail out early if a limit was specified
-                    sorted_revids[:0] = skipped_revids
-                    for revid in sorted_revids:
-                        distances[revid] = len(distances)
-                    break
-                # all parents will need to be pushed as soon as possible
-                for parent in parent_ids_of[revision]:
-                    if parent not in pending_ids:
-                        pending_ids.insert(0, parent)
-                if not pending_ids:
-                    break
-                expected_id = pending_ids.pop(0)
-                # if the next expected revid has already been skipped, requeue
-                # the skipped ids, except those that would go right back to the
-                # skipped list.
-                if expected_id in skipped_revids:
-                    pos = skipped_revids.index(expected_id)
-                    sorted_revids[:0] = skipped_revids[pos:]
-                    del skipped_revids[pos:]
-        self.distances = distances
-        return sorted(distances, key=distances.get)
-
-    def choose_null_colour(self, revid):
-        """We know we don't need this color, so just set it to NULL"""
-        self.colours[revid] = 0
-
-    def choose_colour(self, revid):
-        revision = self.revisions[revid]
-        children_of_id = self.children_of_id
-        parent_ids_of = self.parent_ids_of
-        colours = self.colours
-        # choose colour
-        the_children = children_of_id[revid]
-        if len(the_children) == 1:
-            [child] = the_children
-            if len(parent_ids_of[child]) == 1:
-                # one-one relationship between parent and child, same
-                # colour
-                colours[revid] = colours[child.revision_id]
-            else:
-                self.choose_colour_one_child(revision, child)
-        else:
-            self.choose_colour_many_children(revision, the_children)
-
-    def choose_colour_one_child(self, revision, child):
-        revid = revision.revision_id
-        direct_parent_of = self.direct_parent_of
-        revisions = self.revisions
-        # one child with multiple parents, the first parent with
-        # the same committer gets the colour
-        direct_parent = direct_parent_of.get(child)
-        if direct_parent is None:
-            # if it has not been found yet, find it now and remember
-            for parent_id in self.parent_ids_of[child]:
-                parent_revision = revisions[parent_id]
-                if parent_revision.committer == child.committer:
-                    # found the first parent with the same committer
-                    direct_parent = parent_revision
-                    direct_parent_of[child] = direct_parent
-                    break
-        if direct_parent == revision:
-            self.colours[revid] = self.colours[child.revision_id]
-        else:
-            self.colours[revid] = self.last_colour = self.last_colour + 1
-
-    def choose_colour_many_children(self, revision, the_children):
-        """Colour revision revision."""
-        revid = revision.revision_id
-        direct_parent_of = self.direct_parent_of
-        # multiple children, get the colour of the last displayed child
-        # with the same committer which does not already have its colour
-        # taken
-        available = {}
-        for child in the_children:
-            if child.committer != revision.committer:
-                continue
-            direct_parent = direct_parent_of.get(child)
-            if direct_parent == revision:
-                self.colours[revid] = self.colours[child.revision_id]
-                break
-            # FIXME: Colouring based on whats been displayed MUST be done with 
-            # knowledge of the revisions being output.
-            # until the refactoring to fold graph() into this more compactly is
-            # done, I've disabled this reuse. RBC 20060403
-            # if direct_parent is None:
-            #     available[child] = distances[child.revision_id] 
-            #   .. it will be something like available[child] =  \
-            #  revs[child.revision_id][0] - which is the sequence number
-        else:
-            if available:
-                sorted_children = sorted(available, key=available.get)
-                child = sorted_children[-1]
-                direct_parent_of[child] = revision
-                self.colours[revid] = self.colours[child.revision_id]
-            else:
-                # no candidate children is available, pick the next
-                # colour
-                self.colours[revid] = self.last_colour = self.last_colour + 1
-
-
-def distances(repository, start_revid, maxnum=None):
-    """Sort the revisions.
-
-    Traverses the branch revision tree starting at start and produces an
-    ordered list of revisions such that a revision always comes after
-    any revision it is the parent of.
-
-    Returns a tuple of (revids, revisions, colours, children)
-    """
-    distance = DistanceMethod(repository, start_revid)
-    distance.fill_caches()
-    distance.merge_sorted = merge_sort(distance.graph, distance.start_revid)
-    children = distance.make_children_map()
-    
-    count = 0
-    for seq, revid, merge_depth, end_of_merge in distance.merge_sorted:
-        count += 1
-        if maxnum is not None and count > maxnum:
-            distance.choose_null_colour(revid)
-        else:
-            distance.choose_colour(revid)
-
-    revisions = distance.revisions
-    colours = distance.colours
-    parent_ids_of = distance.parent_ids_of
-    return (revisions, colours, children, parent_ids_of, distance.merge_sorted)
-
-
-def graph(revisions, colours, merge_sorted):
-    """Produce a directed graph of a bzr branch.
-
-    For each revision it then yields a tuple of (revision, node, lines).
-    If the revision is only referenced in the branch and not present in the
-    store, revision will be a DummyRevision object, otherwise it is the bzr
-    Revision object with the meta-data for the revision.
-
-    Node is a tuple of (column, colour) with column being a zero-indexed
-    column number of the graph that this revision represents and colour
-    being a zero-indexed colour (which doesn't specify any actual colour
-    in particular) to draw the node in.
-
-    Lines is a list of tuples which represent lines you should draw away
-    from the revision, if you also need to draw lines into the revision
-    you should use the lines list from the previous iteration.  Each
-    typle in the list is in the form (start, end, colour) with start and
-    end being zero-indexed column numbers and colour as in node.
-
-    It's up to you how to actually draw the nodes and lines (straight,
-    curved, kinked, etc.) and to pick the actual colours for each index.
-    """
-    if not len(merge_sorted):
-        return
-    # split merge_sorted into a map:
-    revs = {}
-    # FIXME: get a hint on this from the merge_sorted data rather than
-    # calculating it ourselves
-    # mapping from rev_id to the sequence number of the next lowest rev
-    next_lower_rev = {}
-    # mapping from rev_id to next-in-branch-revid - may be None for end
-    # of branch
-    next_branch_revid = {}
-    # the stack we are in in the sorted data for determining which 
-    # next_lower_rev to set. It is a stack which has one list at each
-    # depth - the ids at that depth that need the same id allocated.
-    current_stack = [[]]
-    for seq, revid, indent, end_merge in merge_sorted:
-        revs[revid] = (seq, indent, end_merge)
-        if indent == len(current_stack):
-            # new merge group starts
-            current_stack.append([revid])
-        elif indent == len(current_stack) - 1:
-            # part of the current merge group
-            current_stack[-1].append(revid)
-        else:
-            # end of a merge group
-            while current_stack[-1]:
-                stack_rev_id = current_stack[-1].pop()
-                # record the next lower rev for this rev:
-                next_lower_rev[stack_rev_id] = seq
-                # if this followed a non-end-merge rev in this group note that
-                if len(current_stack[-1]):
-                    if not revs[current_stack[-1][-1]][2]:
-                        next_branch_revid[current_stack[-1][-1]] = stack_rev_id
-            current_stack.pop()
-            # append to the now-current merge group
-            current_stack[-1].append(revid)
-    # assign a value to all the depth 0 revisions
-    while current_stack[-1]:
-        stack_rev_id = current_stack[-1].pop()
-        # record the next lower rev for this rev:
-        next_lower_rev[stack_rev_id] = len(merge_sorted)
-        # if this followed a non-end-merge rev in this group note that
-        if len(current_stack[-1]):
-            if not revs[current_stack[-1][-1]][2]:
-                next_branch_revid[current_stack[-1][-1]] = stack_rev_id
-
-    # a list of the current revisions we are drawing lines TO indicating
-    # the sequence of their lines on the screen.
-    # i.e. [A, B, C] means that the line to A, to B, and to C are in
-    # (respectively), 0, 1, 2 on the screen.
-    hanging = [merge_sorted[0][1]]
-    for seq, revid, indent, end_merge in merge_sorted:
-        # a list of the lines to draw: their position in the
-        # previous row, their position in this row, and the colour
-        # (which is the colour they are routing to).
-        lines = []
-
-        new_hanging = []
-
-        for h_idx, hang in enumerate(hanging):
-            # one of these will be the current lines node:
-            # we are drawing a line. h_idx 
-            if hang == revid:
-                # we have found the current lines node
-                node = (h_idx, colours[revid])
-
-                # note that we might have done the main parent
-                drawn_parents = set()
-
-                def draw_line(from_idx, to_idx, revision_id):
-                    try:
-                        n_idx = new_hanging.index(revision_id)
-                    except ValueError:
-                        # force this to be vertical at the place this rev was
-                        # drawn.
-                        new_hanging.insert(to_idx, revision_id)
-                        n_idx = to_idx
-                    lines.append((from_idx, n_idx, colours[revision_id]))
-
-                
-                # we want to draw a line to the next commit on 'this' branch
-                if not end_merge:
-                    # drop this line first.
-                    parent_id = next_branch_revid[revid]
-                    draw_line(h_idx, h_idx, parent_id)
-                    # we have drawn this parent
-                    drawn_parents.add(parent_id)
-                else:
-                    # this is the last revision in a 'merge', show where it came from
-                    if len(revisions[revid].parent_ids) > 1:
-                        # having > 1
-                        # parents means this commit was a merge, and being
-                        # the end point of a merge group means that all
-                        # the parent revisions were merged into branches
-                        # to the left of this before this was committed
-                        # - so we want to show this as a new branch from
-                        # those revisions.
-                        # to do this, we show the parent with the lowest
-                        # sequence number, which is the one that this
-                        # branch 'spawned from', and no others.
-                        # If this sounds like a problem, remember that:
-                        # if the parent was not already in our mainline
-                        # it would show up as a merge into this making
-                        # this not the end of a merge-line.
-                        lowest = len(merge_sorted)
-                        for parent_id in revisions[revid].parent_ids:
-                            if revs[parent_id][0] < lowest:
-                                lowest = revs[parent_id][0]
-                        assert lowest != len(merge_sorted)
-                        draw_line(h_idx, len(new_hanging), merge_sorted[lowest][1])
-                        drawn_parents.add(merge_sorted[lowest][1])
-                    elif len(revisions[revid].parent_ids) == 1:
-                        # only one parent, must show this link to be useful.
-                        parent_id = revisions[revid].parent_ids[0]
-                        draw_line(h_idx, len(new_hanging), parent_id)
-                        drawn_parents.add(parent_id)
-                
-                # what do we want to draw lines to from here:
-                # each parent IF its relevant.
-                #
-                # Now we need to hang its parents, we put them at the point
-                # the old column was so anything to the right of this has
-                # to move outwards to make room.  We also try and collapse
-                # hangs to keep the graph small.
-                # RBC: we do not draw lines to parents that were already merged
-                # unless its the last revision in a merge group.
-                for parent_id in revisions[revid].parent_ids:
-                    if parent_id in drawn_parents:
-                        continue
-                    parent_seq = revs[parent_id][0]
-                    parent_depth = revs[parent_id][1]
-                    if parent_depth == indent + 1:
-                        # The parent was a merge into this branch determine if
-                        # it was already merged into the mainline via a
-                        # different merge: if all revisions between us and
-                        # parent_seq have a indent greater than there are no
-                        # revisions with a lower indent than us.
-                        # We do not use 'parent_depth < indent' because that
-                        # would allow un-uniqueified merges to show up, and
-                        # merge_sorted should take care of that for us (but
-                        # does not trim the values)
-                        if parent_seq < next_lower_rev[revid]:
-                            draw_line(h_idx, len(new_hanging), parent_id)
-                    elif parent_depth == indent and parent_seq == seq + 1:
-                        # part of this branch
-                        draw_line(h_idx, len(new_hanging), parent_id)
-            else:
-                # draw a line from the previous position of this line to the 
-                # new position.
-                # h_idx is the old position.
-                # new_indent is the new position. 
-                draw_line(h_idx, len(new_hanging), hang)
-        # we've calculated the row, assign new_hanging to hanging to setup for
-        # the next row
-        hanging = new_hanging
-
-        yield (revisions[revid], node, lines)
-
-
-def same_branch(a, b):
-    """Return whether we think revisions a and b are on the same branch."""
-    if len(a.parent_ids) == 1:
-        # Defacto same branch if only parent
-        return True
-    elif a.committer == b.committer:
-        # Same committer so may as well be
-        return True
-    else:
-        return False
+# -*- coding: UTF-8 -*-
+"""Directed graph production.
+
+This module contains the code to produce an ordered directed graph of a
+bzr branch, such as we display in the tree view at the top of the bzrk
+window.
+"""
+
+__copyright__ = "Copyright © 2005 Canonical Ltd."
+__author__    = "Scott James Remnant <scott@ubuntu.com>"
+
+from bzrlib.tsort import merge_sort
+
+def linegraph(branch, start, maxnum):
+    """Produce a directed graph of a bzr branch.
+
+    Returns a tuple of (line_graph, revid_index, columns_len) where
+    * line_graph is a list of tuples of (revid,
+                                         node,
+                                         lines,
+                                         parents,
+                                         children,
+                                         revno_sequence),
+    * revid_index is a dict of each revision with the key being the revid, and
+      the value the row index, and
+    * columns_len is the number of columns need to draw the line graph.
+    
+
+    Node is a tuple of (column, colour) with column being a zero-indexed
+    column number of the graph that this revision represents and colour
+    being a zero-indexed colour (which doesn't specify any actual colour
+    in particular) to draw the node in.
+
+    Lines is a list of tuples which represent lines you should draw away
+    from the revision, if you also need to draw lines into the revision
+    you should use the lines list from the previous iteration.  Each
+    typle in the list is in the form (start, end, colour) with start and
+    end being zero-indexed column numbers and colour as in node.
+
+    It's up to you how to actually draw the nodes and lines (straight,
+    curved, kinked, etc.) and to pick the actual colours for each index.
+    """
+    
+    # FIXME: This should be configurable
+    BROKEN_LINE_LENGTH = 32
+    
+    # We get the mainline so we can pass it to merge_sort to make merge_sort
+    # run faster.
+    mainline = branch.revision_history()
+    graph_parents = branch.repository.get_revision_graph(start)
+    graph_children = {}
+    for revid in graph_parents.iterkeys():
+        graph_children[revid] = []
+
+    merge_sorted_revisions = merge_sort(
+        graph_parents,
+        start,
+        mainline,
+        generate_revno=True)
+    
+    revid_index = {}
+    revno_index = {}
+    
+    # This will hold an item for each "branch". For a revisions, the revsion
+    # number less the least significant digit is the branch_id, and used as the
+    # key for the dict. Hence revision with the same revsion number less the
+    # least significant digit are considered to be in the same branch line.
+    # e.g.: for revisions 290.12.1 and 290.12.2, the branch_id would be 290.12,
+    # and these two revisions will be in the same branch line. Each value is
+    # a list of rev_indexes in the branch.
+    branch_lines = {}
+    
+    linegraph = []    
+    
+    for (rev_index, (sequence_number,
+                     revid,
+                     merge_depth,
+                     revno_sequence,
+                     end_of_merge)) in enumerate(merge_sorted_revisions):
+        if maxnum and rev_index >= maxnum:
+            break
+        revid_index[revid] = rev_index
+        revno_index[revno_sequence] = rev_index
+        
+        branch_id = revno_sequence[0:-1]
+        
+        branch_line = None
+        if branch_id not in branch_lines:
+            branch_line = []
+            branch_lines[branch_id] = branch_line
+        else:
+            branch_line = branch_lines[branch_id]
+        
+        branch_line.append(rev_index)
+        
+        parents = graph_parents[revid]
+        for parent_revid in parents:
+            graph_children[parent_revid].append(revid)
+        
+        linegraph.append([revid,
+                          None,
+                          [],
+                          parents,
+                          None,
+                          revno_sequence])
+
+    branch_ids = branch_lines.keys()
+    
+    def branch_id_cmp(x, y):
+        """Compaire branch_id's first by the number of digits, then reversed
+        by their value"""
+        len_x = len(x)
+        len_y = len(y)
+        if len_x == len_y:
+            return -cmp(x, y)
+        return cmp(len_x, len_y)
+    
+    branch_ids.sort(branch_id_cmp)
+    # This will hold a tuple of (child_index, parent_index, col_index) for each
+    # line that needs to be drawn. If col_index is not none, then the line is
+    # drawn along that column, else the the line can be drawn directly between
+    # the child and parent because either the child and parent are in the same
+    # branch line, or the child and parent are 1 row apart.
+    lines = []
+    empty_column = [False for i in range(len(graph_parents))]
+    # This will hold a bit map for each cell. If the cell is true, then the
+    # cell allready contains a node or line. This use when deciding what column
+    # to place a branch line or line in, without it overlaping something else.
+    columns = [list(empty_column)]
+    
+    
+    for branch_id in branch_ids:
+        branch_line = branch_lines[branch_id]
+        
+        # Find the col_index for the direct parent branch. This will be the
+        # starting point when looking for a free column.
+        parent_col_index = 0
+        parent_index = None
+        if len(branch_id) > 1:
+            parent_revno = branch_id[0:-1]
+            if parent_revno in revno_index:
+                parent_index = revno_index[parent_revno]
+                parent_node = linegraph[parent_index][1]
+                if parent_node:
+                    parent_col_index = parent_node[0]
+                
+        
+        col_search_order = _branch_line_col_search_order(columns,
+                                                         parent_col_index)
+        color = reduce(lambda x, y: x+y, branch_id, 0)
+        cur_cont_line = []
+        
+        line_range = []
+        last_rev_index = None
+        for rev_index in branch_line:
+            if last_rev_index:
+                if BROKEN_LINE_LENGTH and \
+                   rev_index - last_rev_index > BROKEN_LINE_LENGTH:
+                    line_range.append(last_rev_index+1)
+                    line_range.append(rev_index-1)
+                else:
+                    line_range.extend(range(last_rev_index+1, rev_index))
+            
+            line_range.append(rev_index)
+            last_rev_index = rev_index
+        
+        if parent_index:
+            if BROKEN_LINE_LENGTH and \
+               parent_index - last_rev_index > BROKEN_LINE_LENGTH:
+                line_range.append(last_rev_index+1)
+            else:
+                line_range.extend(range(last_rev_index+1, parent_index))
+        
+        col_index = _find_free_column(columns,
+                                      empty_column,
+                                      col_search_order,
+                                      line_range)
+        node = (col_index, color)
+        for rev_index in branch_line:
+            linegraph[rev_index][1] = node
+            columns[col_index][rev_index] = True
+        
+        for rev_index in branch_line:
+            (sequence_number,
+                 revid,
+                 merge_depth,
+                 revno_sequence,
+                 end_of_merge) = merge_sorted_revisions[rev_index]
+            
+            linegraph[rev_index][4] = graph_children[revid]
+            col_index = linegraph[rev_index][1][0]
+            
+            for parent_revid in graph_parents[revid]:
+                if parent_revid in revid_index:
+                    
+                    parent_index = revid_index[parent_revid]                            
+                    parent_node = linegraph[parent_index][1]
+                    if parent_node:
+                        parent_col_index = parent_node[0]
+                    else:
+                        parent_col_index = None
+                    col_search_order = \
+                            _line_col_search_order(columns,
+                                                   parent_col_index,
+                                                   col_index)
+                        
+                    # If this line is really long, break it.
+                    if len(branch_id) > 0 and \
+                       BROKEN_LINE_LENGTH and \
+                       parent_index - rev_index > BROKEN_LINE_LENGTH:
+                        child_line_col_index = \
+                            _find_free_column(columns,
+                                              empty_column,
+                                              col_search_order,
+                                              (rev_index + 1,))
+                        _mark_column_as_used(columns,
+                                             child_line_col_index,
+                                             (rev_index + 1,))
+                        
+                        # Recall _line_col_search_order to reset it back to
+                        # the beging.
+                        col_search_order = \
+                                _line_col_search_order(columns,
+                                                       parent_col_index,
+                                                       col_index)
+                        parent_col_line_index = \
+                            _find_free_column(columns,
+                                              empty_column,
+                                              col_search_order,
+                                              (parent_index - 1,))
+                        _mark_column_as_used(columns,
+                                             parent_col_line_index,
+                                             (parent_index - 1,))
+                        lines.append((rev_index,
+                                      parent_index,
+                                      (child_line_col_index,
+                                       parent_col_line_index)))
+                    else :
+                        line_col_index = col_index
+                        if parent_index - rev_index >1:
+                            line_range = range(rev_index + 1, parent_index)
+                            line_col_index = \
+                                _find_free_column(columns,
+                                                  empty_column,
+                                                  col_search_order,
+                                                  line_range)
+                            _mark_column_as_used(columns,
+                                                 line_col_index,
+                                                 line_range)
+                        lines.append((rev_index,
+                                      parent_index,
+                                      (line_col_index,)))
+    
+    for (child_index, parent_index, line_col_indexes) in lines:
+        (child_col_index, child_color) = linegraph[child_index][1]
+        (parent_col_index, parent_color) = linegraph[parent_index][1]
+        
+        if len(line_col_indexes) == 1:
+            if parent_index - child_index == 1:
+                linegraph[child_index][2].append(
+                    (child_col_index,
+                     parent_col_index,
+                     parent_color))
+            else:
+                # line from the child's column to the lines column
+                linegraph[child_index][2].append(
+                    (child_col_index,
+                     line_col_indexes[0],
+                     parent_color))
+                # lines down the line's column
+                for line_part_index in range(child_index+1, parent_index-1):
+                    linegraph[line_part_index][2].append(
+                        (line_col_indexes[0],   
+                         line_col_indexes[0],
+                         parent_color))
+                # line from the line's column to the parent's column
+                linegraph[parent_index-1][2].append(
+                    (line_col_indexes[0],
+                     parent_col_index,
+                     parent_color))
+        else:
+            # Broken line
+            # line from the child's column to the lines column
+            linegraph[child_index][2].append(
+                (child_col_index,
+                 line_col_indexes[0],
+                 parent_color))
+            # Broken line end
+            linegraph[child_index+1][2].append(
+                (line_col_indexes[0],
+                 None,
+                 parent_color))
+            
+            # Broken line end 
+            linegraph[parent_index-2][2].append(
+                (None,
+                 line_col_indexes[1],
+                 parent_color))
+            # line from the line's column to the parent's column
+            linegraph[parent_index-1][2].append(
+                (line_col_indexes[1],
+                 parent_col_index,
+                 parent_color))
+            
+    
+    return (linegraph, revid_index, len(columns))
+
+def _branch_line_col_search_order(columns, parent_col_index):
+    for col_index in range(parent_col_index, len(columns)):
+        yield col_index
+    for col_index in range(parent_col_index-1, -1, -1):
+        yield col_index
+
+def _line_col_search_order(columns, parent_col_index, child_col_index):
+    if parent_col_index is not None:
+        max_index = max(parent_col_index, child_col_index)
+        min_index = min(parent_col_index, child_col_index)
+        for col_index in range(max_index, min_index -1, -1):
+            yield col_index
+    else:
+        max_index = child_col_index
+        min_index = child_col_index
+        yield child_col_index
+    i = 1
+    while max_index + i < len(columns) or \
+          min_index - i > -1:
+        if max_index + i < len(columns):
+            yield max_index + i
+        if min_index - i > -1:
+            yield min_index - i
+        i += 1
+
+def _find_free_column(columns, empty_column, col_search_order, line_range):
+    for col_index in col_search_order:
+        column = columns[col_index]
+        has_overlaping_line = False
+        for row_index in line_range:
+            if column[row_index]:
+                has_overlaping_line = True
+                break
+        if not has_overlaping_line:
+            break
+    else:
+        col_index = len(columns)
+        column = list(empty_column)
+        columns.append(column)
+    return col_index
+
+def _mark_column_as_used(columns, col_index, line_range):
+    column = columns[col_index]
+    for row_index in line_range:
+        column[row_index] = True    
+
+def same_branch(a, b):
+    """Return whether we think revisions a and b are on the same branch."""
+    if len(a.parent_ids) == 1:
+        # Defacto same branch if only parent
+        return True
+    elif a.committer == b.committer:
+        # Same committer so may as well be
+        return True
+    else:
+        return False

=== added file 'viz/treemodel.py'
--- viz/treemodel.py	1970-01-01 00:00:00 +0000
+++ viz/treemodel.py	2007-09-20 19:42:26 +0000
@@ -0,0 +1,110 @@
+# -*- coding: UTF-8 -*-
+"""Tree model.
+
+"""
+
+__copyright__ = "Copyright  2005 Canonical Ltd."
+__author__    = "Gary van der Merwe <garyvdm@gmail.com>"
+
+
+import gtk
+import gobject
+import pango
+
+from bzrlib.osutils import format_date
+
+REVID = 0
+NODE = 1
+LINES = 2
+LAST_LINES = 3
+REVNO = 4
+MESSAGE = 5
+COMMITER = 6
+TIMESTAMP = 7
+REVISION = 8
+PARENTS = 9
+CHILDREN = 10
+
+class TreeModel(gtk.GenericTreeModel):
+
+    
+    def __init__ (self, branch, line_graph_data):
+        gtk.GenericTreeModel.__init__(self)
+        self.revisions = {}
+        self.branch = branch
+        self.line_graph_data = line_graph_data
+    
+    def on_get_flags(self):
+        return gtk.TREE_MODEL_LIST_ONLY
+    
+    def on_get_n_columns(self):
+        return 11
+    
+    def on_get_column_type(self, index):
+        if index == REVID: return gobject.TYPE_STRING
+        if index == NODE: return gobject.TYPE_PYOBJECT
+        if index == LINES: return gobject.TYPE_PYOBJECT
+        if index == LAST_LINES: return gobject.TYPE_PYOBJECT
+        if index == REVNO: return gobject.TYPE_STRING
+        if index == MESSAGE: return gobject.TYPE_STRING
+        if index == COMMITER: return gobject.TYPE_STRING
+        if index == TIMESTAMP: return gobject.TYPE_STRING
+        if index == REVISION: return gobject.TYPE_PYOBJECT
+        if index == PARENTS: return gobject.TYPE_PYOBJECT
+        if index == CHILDREN: return gobject.TYPE_PYOBJECT
+        
+    def on_get_iter(self, path):
+        return path[0]
+    
+    def on_get_path(self, rowref):
+        return rowref
+    
+    def on_get_value(self, rowref, column):
+        (revid, node, lines, parents,
+         children, revno_sequence) = self.line_graph_data[rowref]
+        if column == REVID: return revid
+        if column == NODE: return node
+        if column == LINES: return lines
+        if column == PARENTS: return parents
+        if column == CHILDREN: return children
+        if column == LAST_LINES:
+            if rowref>0:
+                return self.line_graph_data[rowref-1][2]
+            return []
+        if column == REVNO: return ".".join(["%d" % (revno)
+                                      for revno in revno_sequence])
+        
+        if revid not in self.revisions:
+            revision = self.branch.repository.get_revisions([revid])[0]
+            self.revisions[revid] = revision
+        else:
+            revision = self.revisions[revid]
+        
+        if column == REVISION: return revision
+        if column == MESSAGE: return revision.message.split("\n")[0]
+        if column == COMMITER: return revision.committer
+        if column == TIMESTAMP: return format_date(revision.timestamp,
+                                                   revision.timezone)
+    
+    def on_iter_next(self, rowref):
+        if rowref < len(self.line_graph_data) - 1:
+            return rowref+1
+        return None
+    
+    def on_iter_children(self, parent):
+        if parent is None: return 0
+        return None
+    
+    def on_iter_has_child(self, rowref):
+        return False
+    
+    def on_iter_n_children(self, rowref):
+        if rowref is None: return len(self.line_graph_data)
+        return 0
+    
+    def on_iter_nth_child(self, parent, n):
+        if parent is None: return n
+        return None
+    
+    def on_iter_parent(self, child):
+        return None

